@@ -540,9 +540,26 @@ class EventManager:
             msg = 'WARNING: Exception when getting script for event.'
             debug.println(debug.LEVEL_WARNING, msg, True)
         else:
-            msg = 'EVENT MANAGER: Getting script for %s' % app
+            skipCheck = [
+                "object:active-descendant-changed",
+                "object:children-changed",
+                "object:column-reordered",
+                "object:row-reordered",
+                "object:property-change",
+                "object:selection-changed"
+                "object:state-changed:checked",
+                "object:state-changed:expanded",
+                "object:state-changed:indeterminate",
+                "object:state-changed:pressed",
+                "object:state-changed:selected",
+                "object:state-changed:sensitive",
+                "object:state-changed:showing",
+                "object:text-changed",
+            ]
+            check = not list(filter(lambda x: event.type.startswith(x), skipCheck))
+            msg = 'EVENT MANAGER: Getting script for %s (check: %s)' % (app, check)
             debug.println(debug.LEVEL_INFO, msg, True)
-            script = _scriptManager.getScript(app, event.source)
+            script = _scriptManager.getScript(app, event.source, sanityCheck=check)
 
         msg = 'EVENT MANAGER: Script is %s' % script
         debug.println(debug.LEVEL_INFO, msg, True)
@@ -591,6 +608,10 @@ class EventManager:
                and event.detail1):
             return True, "Event source claimed focus."
 
+        if eType.startswith('object:state-changed:selected') and event.detail1 \
+           and role == pyatspi.ROLE_MENU and state.contains(pyatspi.STATE_FOCUSED):
+            return True, "Selection change in focused menu"
+
         # This condition appears with gnome-screensave-dialog.
         # See bug 530368.
         if eType.startswith('object:state-changed:showing') \
@@ -623,6 +644,18 @@ class EventManager:
                 debug.printException(debug.LEVEL_WARNING)
                 return
 
+        if eType.startswith("window"):
+            _scriptManager.reclaimScripts()
+
+        if eType.startswith("object:state-changed:active"):
+            try:
+                role = event.source.getRole()
+            except:
+                pass
+            else:
+                if role == pyatspi.ROLE_FRAME:
+                    _scriptManager.reclaimScripts()
+
         # Clean up any flat review context so that Orca does not get
         # confused (see bgo#609633)
         #
