@@ -1,6 +1,7 @@
 # Orca
 #
 # Copyright 2009 Sun Microsystems Inc.
+# Copyright 2015-2016 Igalia, S.L.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -22,22 +23,23 @@
 __id__        = "$Id:$"
 __version__   = "$Revision:$"
 __date__      = "$Date:$"
-__copyright__ = "Copyright (c) 2009 Sun Microsystems Inc."
+__copyright__ = "Copyright (c) 2009 Sun Microsystems Inc." \
+                "Copyright (c) 2015-2016 Igalia, S.L."
 __license__   = "LGPL"
 
+import collections
+import pyatspi
 import sys
 import time
 import traceback
-
-import pyatspi
+from gi.repository import Atspi, Atk
 
 from . import braille
 from . import debug
 from . import messages
 from . import object_properties
 from . import settings
-
-import collections
+from . import settings_manager
 
 def _formatExceptionInfo(maxTBlevel=5):
     cla, exc, trbk = sys.exc_info()
@@ -58,6 +60,8 @@ def _formatExceptionInfo(maxTBlevel=5):
 #
 METHOD_PREFIX = "_generate"
 
+_settingsManager = settings_manager.getManager()
+
 class Generator:
     """Takes accessible objects and generates a presentation for those
     objects.  See the generate method, which is the primary entry
@@ -71,6 +75,7 @@ class Generator:
 
         self._mode = mode
         self._script = script
+        self._activeProgressBars = {}
         self._methodsDict = {}
         for method in \
             [z for z in [getattr(self, y).__get__(self, self.__class__) for y in [x for x in dir(self) if x.startswith(METHOD_PREFIX)]] if isinstance(z, collections.Callable)]:
@@ -151,6 +156,9 @@ class Generator:
     def generateContents(self, contents, **args):
         return []
 
+    def generateContext(self, obj, **args):
+        return []
+
     def generate(self, obj, **args):
         """Returns an array of strings (and possibly voice and audio
         specifications) that represent the complete presentatin for the
@@ -253,6 +261,9 @@ class Generator:
         for element in result:
             debug.println(debug.LEVEL_ALL, "           %s" % element)
 
+        if args.get('isProgressBarUpdate') and result:
+            self.setProgressBarUpdateTimeAndValue(obj)
+
         return result
 
     #####################################################################
@@ -352,7 +363,7 @@ class Generator:
         result.extend(label)
         if not len(label):
             result.extend(name)
-        elif len(name) and name[0].strip() != label[0].strip():
+        elif len(name) and name[0].split() != label[0].split():
             result.extend(name)
         return result
 
@@ -379,21 +390,20 @@ class Generator:
 
         role = args.get('role', obj.getRole())
 
-        # The text in the description is the same as the text in the page
-        # tab and similar to (and sometimes the same as) the prompt.
-        if role == pyatspi.ROLE_TERMINAL:
-            return []
-
         # Unity Panel Service menubar items are labels which claim focus and
         # have an accessible description of the text + underscore symbol used
         # to create the mnemonic. We'll work around that here for now.
         if role == pyatspi.ROLE_LABEL:
             return []
 
+        if role == pyatspi.ROLE_ICON:
+            name = self._script.utilities.displayedText(obj) or ""
+        else:
+            name = obj.name or ""
+
         result = []
         if obj.description:
             label = self._script.utilities.displayedLabel(obj) or ""
-            name = obj.name or ""
             desc = obj.description.lower()
             if not (desc in name.lower() or desc in label.lower()):
                 result.append(obj.description)
@@ -459,6 +469,25 @@ class Generator:
             result.append(self._script.formatting.getString(**args))
         return result
 
+    def _generateInvalid(self, obj, **args):
+        error = self._script.utilities.getError(obj)
+        if not error:
+            return []
+
+        result = []
+        if not args.get('mode', None):
+            args['mode'] = self._mode
+        args['stringType'] = 'invalid'
+        indicators = self._script.formatting.getString(**args)
+
+        if error == 'spelling':
+            result.append(indicators[1])
+        elif error == 'grammar':
+            result.append(indicators[2])
+        else:
+            result.append(indicators[0])
+        return result
+
     def _generateRequired(self, obj, **args):
         """Returns an array of strings for use by speech and braille that
         represent the required state of the object, but only if it is
@@ -558,6 +587,20 @@ class Generator:
 
         return []
 
+    def _generateSwitchState(self, obj, **args):
+        result = []
+        if not args.get('mode', None):
+            args['mode'] = self._mode
+        args['stringType'] = 'switch'
+        indicators = self._script.formatting.getString(**args)
+        state = obj.getState()
+        if state.contains(pyatspi.STATE_CHECKED) \
+           or state.contains(pyatspi.STATE_PRESSED):
+            result.append(indicators[1])
+        else:
+            result.append(indicators[0])
+        return result
+
     def _generateToggleState(self, obj, **args):
         """Returns an array of strings for use by speech and braille that
         represent the checked state of the object.  This is typically
@@ -646,7 +689,7 @@ class Generator:
         if not text:
             return result
 
-        roleString =  self.getLocalizedRoleName(obj, pyatspi.ROLE_ROW_HEADER)
+        roleString =  self.getLocalizedRoleName(obj, role=pyatspi.ROLE_ROW_HEADER)
         if args.get('mode') == 'speech':
             if settings.speechVerbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
                and not args.get('formatType') in ['basicWhereAmI', 'detailedWhereAmI']:
@@ -672,7 +715,7 @@ class Generator:
         if not text:
             return result
 
-        roleString =  self.getLocalizedRoleName(obj, pyatspi.ROLE_COLUMN_HEADER)
+        roleString =  self.getLocalizedRoleName(obj, role=pyatspi.ROLE_COLUMN_HEADER)
         if args.get('mode') == 'speech':
             if settings.speechVerbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
                and not args.get('formatType') in ['basicWhereAmI', 'detailedWhereAmI']:
@@ -877,7 +920,7 @@ class Generator:
         if start is None or end is None:
             return []
 
-        substring = self._script.utilities.substring(obj, start, end)
+        substring = args.get('string', self._script.utilities.substring(obj, start, end))
         if substring and not self._script.EMBEDDED_OBJECT_CHARACTER in substring:
             return [substring]
 
@@ -1070,6 +1113,55 @@ class Generator:
     def _generateProgressBarValue(self, obj, **args):
         return []
 
+    def _getProgressBarUpdateInterval(self):
+        return int(_settingsManager.getSetting('progressBarUpdateInterval'))
+
+    def _shouldPresentProgressBarUpdate(self, obj, **args):
+        percent = self._script.utilities.getValueAsPercent(obj)
+        lastTime, lastValue = self.getProgressBarUpdateTimeAndValue(obj, type=self)
+        if percent == lastValue:
+            return False
+
+        if percent == 100:
+            return True
+
+        interval = int(time.time() - lastTime)
+        return interval >= self._getProgressBarUpdateInterval()
+
+    def _cleanUpCachedProgressBars(self):
+        isValid = lambda x: not (self._script.utilities.isZombie(x) or self._script.utilities.isDead(x))
+        bars = list(filter(isValid, self._activeProgressBars))
+        self._activeProgressBars = {x:self._activeProgressBars.get(x) for x in bars}
+
+    def _getMostRecentProgressBarUpdate(self):
+        self._cleanUpCachedProgressBars()
+        if not self._activeProgressBars.values():
+            return None, 0.0, None
+
+        sortedValues = sorted(self._activeProgressBars.values(), key=lambda x: x[0])
+        prevTime, prevValue = sortedValues[-1]
+        return list(self._activeProgressBars.keys())[-1], prevTime, prevValue
+
+    def getProgressBarNumberAndCount(self, obj):
+        self._cleanUpCachedProgressBars()
+        if not obj in self._activeProgressBars:
+            self._activeProgressBars[obj] = 0.0, None
+
+        thisValue = self.getProgressBarUpdateTimeAndValue(obj)
+        index = list(self._activeProgressBars.values()).index(thisValue)
+        return index + 1, len(self._activeProgressBars)
+
+    def getProgressBarUpdateTimeAndValue(self, obj, **args):
+        if not obj in self._activeProgressBars:
+            self._activeProgressBars[obj] = 0.0, None
+
+        return self._activeProgressBars.get(obj)
+
+    def setProgressBarUpdateTimeAndValue(self, obj, lastTime=None, lastValue=None):
+        lastTime = lastTime or time.time()
+        lastValue = lastValue or self._script.utilities.getValueAsPercent(obj)
+        self._activeProgressBars[obj] = lastTime, lastValue
+
     def _getAlternativeRole(self, obj, **args):
         if self._script.utilities.isMath(obj):
             if self._script.utilities.isMathFraction(obj):
@@ -1092,5 +1184,116 @@ class Generator:
                 return 'ROLE_MATH_TABLE_ROW'
         if self._script.utilities.isStatic(obj):
             return 'ROLE_STATIC'
+        if self._script.utilities.isSwitch(obj):
+            return 'ROLE_SWITCH'
+        if self._script.utilities.isBlockquote(obj):
+            return pyatspi.ROLE_BLOCK_QUOTE
+        if self._script.utilities.isLandmark(obj):
+            return pyatspi.ROLE_LANDMARK
+        if self._script.utilities.isFocusableLabel(obj):
+            return pyatspi.ROLE_LIST_ITEM
 
         return args.get('role', obj.getRole())
+
+    def getLocalizedRoleName(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if "Value" in pyatspi.listInterfaces(obj):
+            state = obj.getState()
+            isVertical = state.contains(pyatspi.STATE_VERTICAL)
+            isHorizontal = state.contains(pyatspi.STATE_HORIZONTAL)
+            isFocused = state.contains(pyatspi.STATE_FOCUSED) \
+                        or args.get('alreadyFocused', False)
+
+            if role == pyatspi.ROLE_SLIDER:
+                if isHorizontal:
+                    return object_properties.ROLE_SLIDER_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SLIDER_VERTICAL
+            elif role == pyatspi.ROLE_SCROLL_BAR:
+                if isHorizontal:
+                    return object_properties.ROLE_SCROLL_BAR_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SCROLL_BAR_VERTICAL
+            elif role == pyatspi.ROLE_SEPARATOR:
+                if isHorizontal:
+                    return object_properties.ROLE_SPLITTER_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SPLITTER_VERTICAL
+            elif role == pyatspi.ROLE_SPLIT_PANE and isFocused:
+                # The splitter has the opposite orientation of the split pane.
+                if isHorizontal:
+                    return object_properties.ROLE_SPLITTER_VERTICAL
+                if isVertical:
+                    return object_properties.ROLE_SPLITTER_HORIZONTAL
+
+        if self._script.utilities.isMenuButton(obj):
+            return object_properties.ROLE_MENU_BUTTON
+
+        if self._script.utilities.isSwitch(obj):
+            return object_properties.ROLE_SWITCH
+
+        if self._script.utilities.isLandmark(obj):
+            if self._script.utilities.isLandmarkBanner(obj):
+                return object_properties.ROLE_LANDMARK_BANNER
+            if self._script.utilities.isLandmarkComplementary(obj):
+                return object_properties.ROLE_LANDMARK_COMPLEMENTARY
+            if self._script.utilities.isLandmarkContentInfo(obj):
+                return object_properties.ROLE_LANDMARK_CONTENTINFO
+            if self._script.utilities.isLandmarkMain(obj):
+                return object_properties.ROLE_LANDMARK_MAIN
+            if self._script.utilities.isLandmarkNavigation(obj):
+                return object_properties.ROLE_LANDMARK_NAVIGATION
+            if self._script.utilities.isLandmarkRegion(obj):
+                return object_properties.ROLE_LANDMARK_REGION
+            if self._script.utilities.isLandmarkSearch(obj):
+                return object_properties.ROLE_LANDMARK_SEARCH
+            if self._script.utilities.isLandmarkForm(obj):
+                role = pyatspi.ROLE_FORM
+
+        if not isinstance(role, (pyatspi.Role, Atspi.Role)):
+            try:
+                return obj.getLocalizedRoleName()
+            except:
+                return ''
+
+        nonlocalized = Atspi.role_get_name(role)
+        atkRole = Atk.role_for_name(nonlocalized)
+        if atkRole == Atk.Role.INVALID and role == pyatspi.ROLE_STATUS_BAR:
+            atkRole = Atk.role_for_name("statusbar")
+
+        return Atk.role_get_localized_name(atkRole)
+
+    def getStateIndicator(self, obj, **args):
+        if self._script.utilities.isSwitch(obj):
+            return self._generateSwitchState(obj, **args)
+
+        role = args.get('role', obj.getRole())
+
+        if role == pyatspi.ROLE_MENU_ITEM:
+            return self._generateMenuItemCheckedState(obj, **args)
+
+        if role in [pyatspi.ROLE_RADIO_BUTTON, pyatspi.ROLE_RADIO_MENU_ITEM]:
+            return self._generateRadioState(obj, **args)
+
+        if role in [pyatspi.ROLE_CHECK_BOX, pyatspi.ROLE_CHECK_MENU_ITEM]:
+            return self._generateCheckedState(obj, **args)
+
+        if role == pyatspi.ROLE_TOGGLE_BUTTON:
+            return self._generateToggleState(obj, **args)
+
+        if role == pyatspi.ROLE_TABLE_CELL:
+            return self._generateCellCheckedState(obj, **args)
+
+        return []
+
+    def getValue(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if role == pyatspi.ROLE_PROGRESS_BAR:
+            return self._generateProgressBarValue(obj, **args)
+
+        if role in [pyatspi.ROLE_SCROLL_BAR, pyatspi.ROLE_SLIDER]:
+            return self._generatePercentage(obj, **args)
+
+        return []
