@@ -254,6 +254,7 @@ class Script(default.Script):
                 structural_navigation.StructuralNavigation.CHUNK,
                 structural_navigation.StructuralNavigation.CLICKABLE,
                 structural_navigation.StructuralNavigation.COMBO_BOX,
+                structural_navigation.StructuralNavigation.CONTAINER,
                 structural_navigation.StructuralNavigation.ENTRY,
                 structural_navigation.StructuralNavigation.FORM_FIELD,
                 structural_navigation.StructuralNavigation.HEADING,
@@ -655,7 +656,11 @@ class Script(default.Script):
         if not _settingsManager.getSetting('rewindAndFastForwardInSayAll'):
             return False
 
-        obj, start, end, string = self._sayAllContents[0]
+        try:
+            obj, start, end, string = self._sayAllContents[0]
+        except IndexError:
+            return False
+
         orca.setLocusOfFocus(None, obj, notifyScript=False)
         self.utilities.setCaretContext(obj, start)
 
@@ -670,7 +675,11 @@ class Script(default.Script):
         if not _settingsManager.getSetting('rewindAndFastForwardInSayAll'):
             return False
 
-        obj, start, end, string = self._sayAllContents[-1]
+        try:
+            obj, start, end, string = self._sayAllContents[-1]
+        except IndexError:
+            return False
+
         orca.setLocusOfFocus(None, obj, notifyScript=False)
         self.utilities.setCaretContext(obj, end)
 
@@ -722,20 +731,35 @@ class Script(default.Script):
         """Returns True if we should use focus mode in obj."""
 
         if self._focusModeIsSticky:
+            msg = "WEB: Using focus mode because focus mode is sticky"
+            debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
         if self._browseModeIsSticky:
+            msg = "WEB: Not using focus mode because browse mode is sticky"
+            debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
         if not _settingsManager.getSetting('structNavTriggersFocusMode') \
            and self._lastCommandWasStructNav:
+            msg = "WEB: Not using focus mode due to struct nav settings"
+            debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
         if not _settingsManager.getSetting('caretNavTriggersFocusMode') \
            and self._lastCommandWasCaretNav:
+            msg = "WEB: Not using focus mode due to caret nav settings"
+            debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
-        return self.utilities.isFocusModeWidget(obj)
+        if self.utilities.isFocusModeWidget(obj):
+            msg = "WEB: Using focus mode because %s is a focus mode widget" % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        msg = "WEB: Not using focus mode for %s due to lack of cause" % obj
+        debug.println(debug.LEVEL_INFO, msg, True)
+        return False
 
     def speakContents(self, contents, **args):
         """Speaks the specified contents."""
@@ -801,7 +825,7 @@ class Script(default.Script):
 
     def presentObject(self, obj, **args):
         priorObj = None
-        if self._lastCommandWasCaretNav:
+        if self._lastCommandWasCaretNav or args.get("includeContext"):
             priorObj, priorOffset = self.utilities.getPriorContext()
 
         offset = args.get("offset", 0)
@@ -1170,10 +1194,10 @@ class Script(default.Script):
             self.utilities.clearCaretContext()
 
         shouldPresent = self.utilities.isShowingOrVisible(event.source)
-        if not _settingsManager.getSetting('onlySpeakDisplayedText'):
+        if not _settingsManager.getSetting('onlySpeakDisplayedText') and shouldPresent:
             if event.detail1:
                 self.presentMessage(messages.PAGE_LOADING_START)
-            elif event.source.name and shouldPresent:
+            elif event.source.name:
                 msg = messages.PAGE_LOADING_END_NAMED % event.source.name
                 self.presentMessage(msg, resetStyles=False)
             else:
@@ -1232,9 +1256,17 @@ class Script(default.Script):
             orca.setLocusOfFocus(event, obj)
             return True
 
-        msg = "WEB: Setting locus of focus to context obj %s (no notification)" % obj
-        debug.println(debug.LEVEL_INFO, msg, True)
-        orca.setLocusOfFocus(event, obj, False)
+        try:
+            focusState = orca_state.locusOfFocus.getState()
+        except:
+            inFocusedObject = False
+        else:
+            inFocusedObject = focusState.contains(pyatspi.STATE_FOCUSED)
+
+        if not inFocusedObject:
+            msg = "WEB: Setting locus of focus to context obj %s (no notification)" % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
+            orca.setLocusOfFocus(event, obj, False)
 
         self.updateBraille(obj)
         if state.contains(pyatspi.STATE_FOCUSABLE) and not self.utilities.isDocument(obj):
@@ -1261,6 +1293,8 @@ class Script(default.Script):
     def onCaretMoved(self, event):
         """Callback for object:text-caret-moved accessibility events."""
 
+        self.utilities.sanityCheckActiveWindow()
+
         if self.utilities.isZombie(event.source):
             msg = "WEB: Event source is Zombie"
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -1478,30 +1512,27 @@ class Script(default.Script):
             return True
 
         obj, offset = self.utilities.getCaretContext(getZombieReplicant=False)
+        msg = "WEB: Context: %s, %i (focus: %s)" % (obj, offset, orca_state.locusOfFocus)
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         if self.utilities.isZombie(obj):
             obj, offset = self.utilities.getCaretContext(getZombieReplicant=True)
-            if obj and self.utilities.isZombie(orca_state.locusOfFocus):
-                orca.setLocusOfFocus(event, obj, notifyScript=False)
+            if not obj:
+                if self._inFocusMode:
+                    msg = "WEB: Not looking for replicant due to focus mode."
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    return False
 
-        contextDocument = self.utilities.getDocumentForObject(obj)
-        if obj and document == contextDocument and self.utilities.isZombie(obj):
-            msg = "WEB: Context obj %s is Zombie." % obj
-            debug.println(debug.LEVEL_INFO, msg, True)
-            if self._inFocusMode:
-                msg = "WEB: Not looking for replicant due to focus mode."
-                debug.println(debug.LEVEL_INFO, msg, True)
-                return False
-
-            replicant = self.utilities.findReplicant(event.source, obj)
-            if replicant:
-                # Refrain from actually touching the replicant by grabbing
-                # focus or setting the caret in it. Doing so will only serve
-                # to anger it.
-                msg = "WEB: Event handled by updating locusOfFocus and context"
-                debug.println(debug.LEVEL_INFO, msg, True)
-                orca.setLocusOfFocus(event, replicant, False)
-                self.utilities.setCaretContext(replicant, offset)
-                return True
+                obj = self.utilities.findReplicant(event.source, obj)
+                if obj:
+                    # Refrain from actually touching the replicant by grabbing
+                    # focus or setting the caret in it. Doing so will only serve
+                    # to anger it.
+                    msg = "WEB: Event handled by updating locusOfFocus and context"
+                    debug.println(debug.LEVEL_INFO, msg, True)
+                    orca.setLocusOfFocus(event, obj, False)
+                    self.utilities.setCaretContext(obj, offset)
+                    return True
 
         childRole = event.any_data.getRole()
         if childRole == pyatspi.ROLE_ALERT:
@@ -1572,10 +1603,14 @@ class Script(default.Script):
             return False
 
         if self.utilities.isTopLevelWebApp(document):
-            msg = "WEB: Event handled: Setting locusOfFocus to event source"
-            debug.println(debug.LEVEL_INFO, msg, True)
-            orca.setLocusOfFocus(event, event.source)
-            return True
+            if self._browseModeIsSticky:
+                msg = "WEB: Web app claimed focus, but browse mode is sticky"
+                debug.println(debug.LEVEL_INFO, msg, True)
+            else:
+                msg = "WEB: Event handled: Setting locusOfFocus to event source"
+                debug.println(debug.LEVEL_INFO, msg, True)
+                orca.setLocusOfFocus(event, event.source)
+                return True
 
         state = event.source.getState()
         if state.contains(pyatspi.STATE_EDITABLE):
@@ -1590,6 +1625,26 @@ class Script(default.Script):
             orca.setLocusOfFocus(event, event.source)
             return True
 
+        obj, offset = self.utilities.getCaretContext()
+        msg = "WEB: Caret context is %s, %i (focus: %s)" \
+              % (obj, offset, orca_state.locusOfFocus)
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+        if not obj or self.utilities.isZombie(obj):
+            msg = "WEB: Clearing context - obj is null or zombie"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            self.utilities.clearCaretContext()
+
+            obj, offset = self.utilities.searchForCaretContext(event.source)
+            if obj:
+                msg = "WEB: Updating focus and context to %s, %i" % (obj, offset)
+                debug.println(debug.LEVEL_INFO, msg, True)
+                orca.setLocusOfFocus(event, obj, False)
+                self.utilities.setCaretContext(obj, offset)
+            else:
+                msg = "WEB: Search for caret context failed"
+                debug.println(debug.LEVEL_INFO, msg, True)
+
         if self._lastCommandWasCaretNav:
             msg = "WEB: Event ignored: Last command was caret nav"
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -1611,13 +1666,6 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
-        obj, offset = self.utilities.getCaretContext(event.source)
-        if obj and self.utilities.isZombie(obj):
-            msg = "WEB: Clearing context - obj is zombie"
-            debug.println(debug.LEVEL_INFO, msg, True)
-            self.utilities.clearCaretContext()
-            obj, offset = self.utilities.getCaretContext(event.source)
-
         if not obj:
             msg = "WEB: Unable to get non-null, non-zombie context object"
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -1959,3 +2007,12 @@ class Script(default.Script):
         self._lastCommandWasStructNav = False
         self._lastCommandWasMouseButton = False
         return False
+
+    def getTransferableAttributes(self):
+        return {"_lastCommandWasCaretNav": self._lastCommandWasCaretNav,
+                "_lastCommandWasStructNav": self._lastCommandWasStructNav,
+                "_lastCommandWasMouseButton": self._lastCommandWasMouseButton,
+                "_inFocusMode": self._inFocusMode,
+                "_focusModeIsSticky": self._focusModeIsSticky,
+                "_browseModeIsSticky": self._browseModeIsSticky,
+        }
